/*-------------------------------------------------------------------------
pam_coda - PAM authentication and password chaning module - Version 0.1
Copyright (C) BORBELY Zoltan, 1999. All rights reserved

Redistribution and use in source and binary forms of Linux-PAM, with
or without modification, are permitted provided that the following
conditions are met:

1. Redistributions of source code must retain any existing copyright
   notice, and this entire permission notice in its entirety,
   including the disclaimer of warranties.

2. Redistributions in binary form must reproduce all prior and current
   copyright notices, this list of conditions, and the following
   disclaimer in the documentation and/or other materials provided
   with the distribution.

3. The name of any author may not be used to endorse or promote
   products derived from this software without their specific prior
   written permission.

ALTERNATIVELY, this product may be distributed under the terms of the
GNU General Public License, in which case the provisions of the GNU
GPL are required INSTEAD OF the above restrictions.  (This clause is
necessary due to a potential conflict between the GNU GPL and the
restrictions contained in a BSD-style copyright.)

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

History:
--------
09/27/1999	0.1	bozo@szivarvanynet.hu	First release

-------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <syslog.h>
#include <pwd.h>

#define	PAM_SM_AUTH
#define	PAM_SM_PASSWORD
#include <security/pam_modules.h>
#ifdef HAVE_PAM_STD_OPTION
#include <security/pam_mod_misc.h>
#else
#include "pam_mod_misc.h"
#endif

#include <auth2.h>
#include <auser.h>
#include <avenus.h>


#define	MODULE_NAME	"pam_coda"
#define CODA_PROMPT	"Coda password: "

struct	config
{
	int	debug;
	int	pam_options;
	char	authsrv[256];
};

static int  get_config( 	pam_handle_t	*pamh,
				struct config 	*cfg,
				int 		argc,
				const char 	*argv[] )
{
	int	i;

	if(cfg == NULL) {
		syslog(LOG_ERR, "config structure uninitialized\n");
		return(1);
	}

	cfg->debug	= 0;
	cfg->pam_options= 0;
	cfg->authsrv[0]	= '\0';
	for(i=0;i<argc;i++)
	{
		pam_std_option(&(cfg->pam_options),argv[i]);
		if ( strcmp(argv[i],"debug") == 0 )
		{
			cfg->debug	= 1;
			continue;
		}
		if ( strncmp(argv[i],"authsrv=",8) == 0 )
		{
			if ( strlen(argv[i]+8) <= sizeof(cfg->authsrv) )
			{
				strcpy(cfg->authsrv,argv[i]+8);
			}
			continue;
		}
	}
	if ( cfg->debug )
	{
		syslog(LOG_DEBUG, "cfg->debug   = %d\n",cfg->debug);
		syslog(LOG_DEBUG, "cfg->pam_options = %d\n",cfg->pam_options);
		syslog(LOG_DEBUG, "cfg->authsrv = %s\n",cfg->authsrv);
	}
	return(0);
}


static int  _authenticate(	pam_handle_t	*pamh,
				struct config	*cfg,
				char		*user,
				char		*pass )
{
	EncryptedSecretToken	sToken;
	ClearToken		cToken;
	RPC2_Handle		RPCid;
	RPC2_Integer		c_uid;
	uid_t			s_uid;
	int			retval;
	struct passwd		*pwd;

	U_InitRPC();

	retval	= U_BindToServer((cfg->authsrv[0]=='\0') ? NULL : cfg->authsrv,
				 AUTH_METHOD_CODAUSERNAME,
				 user,
				 strlen(user)+1,
				 pass,
				 strlen(pass)+1,
				 &RPCid);

	if ( retval!=0 )
	{
		syslog(LOG_NOTICE,"authentication failure (user: %s) error code: %d",user,retval);
		return( PAM_AUTH_ERR );
	}

	if ( (retval=AuthGetTokens(RPCid,sToken,&cToken)) != 0 )
	{
		AuthQuit(RPCid);
		syslog(LOG_ERR,"AuthGetTokens failed");
		return( PAM_AUTH_ERR );
	}

	if ( AuthNameToId(RPCid,user,&c_uid) != 0 )
	{
		syslog(LOG_NOTICE,"User %s doesn't exist in CODA",user);
		return( PAM_USER_UNKNOWN );
	}
	if( cfg->debug ) 
	{
		syslog(LOG_DEBUG,"CODA UID: %d", c_uid);
	}

	RPC2_Unbind(RPCid);

/* 	As this module doesn't attempt to be full-fledged authentication module
	we resolve any client/server UID conflicts in favour of client (e.g. If
	user has different UID on the local machine and in CODA, tockens will be
	sent to venus from local UID
*/
	s_uid	= geteuid();
	if( (pwd = getpwnam(user)) == NULL)
	{
		syslog(LOG_NOTICE,"User %s doen't exist on local machine", user);
		return( PAM_USER_UNKNOWN );
	}

	if( cfg->debug ) 
	{
		syslog(LOG_DEBUG,"Local UID: %d", pw->pw_uid);
	}

	seteuid(pwd->pw_uid);
	retval	= U_SetLocalTokens(0,&cToken,sToken);
	memset(&cToken,0,sizeof(cToken));
	memset(sToken,0,sizeof(sToken));
	seteuid(s_uid);

	if ( retval != 0 )
	{
		syslog(LOG_ERR,"U_SetLocalTokens failed");
		return( PAM_AUTH_ERR );
	}

	return( PAM_SUCCESS );
}


static int  _changepass(pam_handle_t	*pamh,
			struct config	*cfg,
			char		*user,
			char		*pass,
			char		*cuser,
			char		*cpass )
{
	int	retval;

	U_InitRPC();

	retval	= U_ChangePassword((cfg->authsrv[0]=='\0') ? NULL : cfg->authsrv,
				   user,
				   pass,
				   AUTH_METHOD_CODAUSERNAME,
				   cuser,
				   strlen(cuser)+1,
				   cpass,
				   strlen(cpass)+1);

	return( (retval == 0) ? PAM_SUCCESS : PAM_AUTH_ERR );
}


PAM_EXTERN int	pam_sm_authenticate(	pam_handle_t	*pamh,
					int		flags,
					int		argc,
					const char	*argv[] )
{
	int		retval;
	char		*user, *pass;
	struct config	cfg;

	if(get_config(pamh, &cfg, argc, argv)) {
		return(PAM_SERVICE_ERR);
	}

	if ( (retval=pam_get_user(pamh,(const char **)&user,NULL)) != PAM_SUCCESS ) {
		syslog(LOG_ERR,"couldn't determine name of local user");
		return retval;
	}

	if( (retval = pam_get_pass(pamh, (const char **)&pass, CODA_PROMPT, cfg.pam_options)) != PAM_SUCCESS ) {
		syslog(LOG_ERR,"couldn't get user password");
		return retval;
	}

	if ( (user==NULL) || (pass==NULL) )
	{
		syslog(LOG_ERR,"user name or auth token missing");
		return( PAM_AUTH_ERR );
	}

	return( _authenticate(pamh,&cfg, user, pass) );
}


PAM_EXTERN int	pam_sm_setcred(	pam_handle_t	*pamh,
				int		flags,
				int		argc,
				const char	*argv[] )
{
	return( PAM_SUCCESS );
}


PAM_EXTERN int	pam_sm_chauthtok(	pam_handle_t	*pamh,
					int		flags,
					int		argc,
					const char	*argv[] )
{
	int		retval;
	char		*user, *pass, *cpass;
	struct config	cfg;

	get_config(pamh,&cfg,argc,argv);

	if ( !(flags & PAM_UPDATE_AUTHTOK) )
	{
		return( PAM_SUCCESS );
	}

	if ( (retval=pam_get_user(pamh,(const char **)&user,NULL)) != PAM_SUCCESS )
	{
		syslog(LOG_ERR,"couldn't determine name of local user");
		return( PAM_AUTH_ERR );
	}

	if ( (retval=pam_get_item(pamh,PAM_AUTHTOK,(const void **)&pass)) != PAM_SUCCESS )
	{
		syslog(LOG_ERR,"couldn't get authentication token");
		return( PAM_AUTH_ERR );
	}
	if ( pass == NULL )
	{
		syslog(LOG_ERR,"couldn't get authentication token");
		return( PAM_AUTH_ERR );
	}

	if ( (retval=pam_get_item(pamh,PAM_OLDAUTHTOK,(const void **)&cpass)) != PAM_SUCCESS )
	{
		syslog(LOG_ERR,"couldn't get old authentication token");
		return( PAM_AUTH_ERR );
	}
	if ( cpass == NULL )
	{
		syslog(LOG_ERR,"couldn't get old authentication token");
		return( PAM_AUTH_ERR );
	}

	return( _changepass(pamh,&cfg,user,pass,user,cpass) );
}


/* static module data */

#ifdef PAM_STATIC
struct pam_module _pam_coda_modstruct = {
	MODULE_NAME,
	pam_sm_authenticate,
	pam_sm_setcred,
	NULL,
	NULL,
	NULL,
	pam_sm_chauthtok
};
#endif
